package org.cytoscape.paperwing.internal.cytoscape.view;

import java.util.Collection;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.Map;

import org.cytoscape.view.model.View;
import org.cytoscape.view.model.VisualLexicon;
import org.cytoscape.view.model.VisualLexiconNode;
import org.cytoscape.view.model.VisualProperty;

// This class assumes that the VisualProperty's getIdString() method returns a unique
// value amongst the VisualProperty objects
public abstract class VisualPropertyKeeper<S> implements View<S>{
	
	protected Map<VisualProperty<?>, Object> visualProperties;
	protected Map<VisualProperty<?>, Object> directLocks;
	protected Map<VisualProperty<?>, Object> allLocks;
	
	protected VisualLexicon lexicon;
	
	public VisualPropertyKeeper(VisualLexicon lexicon) {
		
		this.visualProperties = new IdentityHashMap<VisualProperty<?>, Object>();
		this.directLocks = new IdentityHashMap<VisualProperty<?>, Object>();
		this.allLocks = new IdentityHashMap<VisualProperty<?>, Object>();
	}
	
	public <T> T getVisualProperty(VisualProperty<T> visualProperty) {
		Object value = directLocks.get(visualProperty);
		if (value != null)
			return (T) value;

		value = allLocks.get(visualProperty);
		if (value != null)
			return (T) value;
		
		value = visualProperties.get(visualProperty);
		if (value != null)
			return (T) value;
		
		return null;
		
		/*
		// Mapped value is null.  Try default
		value = this.getDefaultValue(visualProperty);
		if(value != null)
			return (T) value;
		else
			return visualProperty.getDefault();
		*/
	}

	public <T, V extends T> void setVisualProperty(VisualProperty<? extends T> visualProperty, V value) {
		
		if (visualProperty == null) {
			visualProperties.remove(visualProperty);
		} else {
			visualProperties.put(visualProperty, value);
		}
	}

	@Override
	public <T, V extends T> void setLockedValue(VisualProperty<? extends T> visualProperty,
			V value) {
		
		this.directLocks.put(visualProperty, value);
		this.allLocks.put(visualProperty, value);
		
		// applyVisualProperty(vp, value);
		VisualLexiconNode node = this.lexicon.getVisualLexiconNode(visualProperty);
		propagateLockedVisualProperty(visualProperty, node.getChildren(), value);
	}
	
	private void propagateLockedVisualProperty(final VisualProperty parent, final Collection<VisualLexiconNode> roots, 
			final Object value) {
		final LinkedList<VisualLexiconNode> nodes = new LinkedList<VisualLexiconNode>();
		nodes.addAll(roots);
		
		while (!nodes.isEmpty()) {
			final VisualLexiconNode node = nodes.pop();
			final VisualProperty vp = node.getVisualProperty();
			
			if (!isDirectlyLocked(vp)) {
				if (parent.getClass() == vp.getClass()) { // Preventing ClassCastExceptions
					this.allLocks.put(vp, value);
					// applyVisualProperty(vp, value);
				}
				
				nodes.addAll(node.getChildren());
			}
		}
	}

	@Override
	public boolean isValueLocked(VisualProperty<?> visualProperty) {
		return (this.allLocks.get(visualProperty) != null);
	}

	@Override
	public boolean isDirectlyLocked(VisualProperty<?> visualProperty) {
		return (this.directLocks.get(visualProperty) != null);
	}
	
	@Override
	public boolean isSet(final VisualProperty<?> vp) {
		return visualProperties.get(vp) != null || allLocks.get(vp) != null || getDefaultValue(vp) != null;
	}
	
	protected abstract <T, V extends T> V getDefaultValue(final VisualProperty<T> vp);
	
	@Override
	public void clearValueLock(VisualProperty<?> visualProperty) {
		directLocks.remove(visualProperty);
		
		VisualLexiconNode root = lexicon.getVisualLexiconNode(visualProperty);
		LinkedList<VisualLexiconNode> nodes = new LinkedList<VisualLexiconNode>();
		nodes.add(root);
		while (!nodes.isEmpty()) {
			VisualLexiconNode node = nodes.pop();
			VisualProperty vp = node.getVisualProperty();
			allLocks.remove(vp);
			
			/*
			// Re-apply the regular visual property value
			if (visualProperties.containsKey(vp)) {
				applyVisualProperty(vp, visualProperties.get(vp));
			// TODO else: reset to the visual style default if visualProperties map doesn't contain this vp
			} else {
				// Apply default if necessary.
				final Object newValue = getVisualProperty(vp);
				applyVisualProperty(vp, newValue);
			}
			*/
			
			for (VisualLexiconNode child : node.getChildren()) {
				if (!isDirectlyLocked(child.getVisualProperty())) {
					nodes.add(child);
				}
			}
			nodes.addAll(node.getChildren());
		}
	}
}
