package org.cytoscape.paperwing.internal.task;

import static org.cytoscape.work.ServiceProperties.APPS_MENU;
import static org.cytoscape.work.ServiceProperties.INSERT_SEPARATOR_AFTER;
import static org.cytoscape.work.ServiceProperties.INSERT_SEPARATOR_BEFORE;
import static org.cytoscape.work.ServiceProperties.MENU_GRAVITY;
import static org.cytoscape.work.ServiceProperties.PREFERRED_ACTION;
import static org.cytoscape.work.ServiceProperties.PREFERRED_MENU;
import static org.cytoscape.work.ServiceProperties.TITLE;
import static org.cytoscape.work.ServiceProperties.TOOLTIP;

import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.AbstractAction;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;

import org.cytoscape.application.swing.CyEdgeViewContextMenuFactory;
import org.cytoscape.application.swing.CyMenuItem;
import org.cytoscape.application.swing.CyNetworkViewContextMenuFactory;
import org.cytoscape.application.swing.CyNodeViewContextMenuFactory;
import org.cytoscape.model.CyEdge;
import org.cytoscape.model.CyNetwork;
import org.cytoscape.model.CyNode;
import org.cytoscape.paperwing.internal.data.GraphicsData;
import org.cytoscape.task.EdgeViewTaskFactory;
import org.cytoscape.task.NetworkViewLocationTaskFactory;
import org.cytoscape.task.NetworkViewTaskFactory;
import org.cytoscape.task.NodeViewTaskFactory;
import org.cytoscape.util.swing.GravityTracker;
import org.cytoscape.util.swing.JMenuTracker;
import org.cytoscape.view.model.CyNetworkView;
import org.cytoscape.view.model.View;
import org.cytoscape.view.model.VisualLexicon;
import org.cytoscape.view.model.VisualProperty;
import org.cytoscape.work.TaskFactory;
//import org.cytoscape.work.TaskFactoryPredicate;
import org.cytoscape.work.swing.DialogTaskManager;
import org.cytoscape.work.swing.DynamicSubmenuListener;

/**
 * This class is responsible for creating and populating pop-up menus created when right-clicking the network.
 */
public class PopupMenuCreator {
		
	/** For creating popup menu items for the network, only use task factories that match the preferred action*/
	private static final String NETWORK_PREFFERED_ACTION = "NEW";
	
	private DialogTaskManager taskManager;
	private VisualLexicon visualLexicon;

	private StaticTaskFactoryProvisioner factoryProvisioner;
	
	// Large value to be used for the gravity value of org.cytoscape.util.swing.GravityTracker
	private double largeValue = Double.MAX_VALUE / 2.0;

	public PopupMenuCreator(
			DialogTaskManager taskManager, 
			VisualLexicon visualLexicon) {
		
		this.taskManager = taskManager;
		this.visualLexicon = visualLexicon;
		this.factoryProvisioner = new StaticTaskFactoryProvisioner();
	}
	
	public JPopupMenu createEdgeMenu(View<CyEdge> edgeView, 
			CyNetworkView networkView, 
			VisualLexicon visualLexicon,
			ViewTaskFactoryListener viewTaskFactoryListener) {
		
		Collection<EdgeViewTaskFactory> usableTFs = getPreferredActions(viewTaskFactoryListener.edgeViewTFs, "NEW");
		Collection<CyEdgeViewContextMenuFactory> usableCMFs = getPreferredActions(viewTaskFactoryListener.cyEdgeViewContextMenuFactory, "NEW");
		View<CyEdge> ev = edgeView;
		
		String edgeLabel = networkView.getModel().getRow(ev.getModel()).get("interaction",String.class);
		JPopupMenu menu = createMenu(edgeLabel);
		JMenuTracker tracker = new JMenuTracker(menu);

		for ( EdgeViewTaskFactory evtf : usableTFs ) {
			Object context = null;
			NamedTaskFactory provisioner = factoryProvisioner.createFor(evtf, ev, networkView);
			addMenuItem(ev, menu, provisioner, context, tracker, viewTaskFactoryListener.edgeViewTFs.get(evtf), this.visualLexicon);
		}
		
		for (CyEdgeViewContextMenuFactory edgeCMF: usableCMFs) {
			// menu.add(edgeCMF.createMenuItem(m_view , ev).getMenuItem());
			CyMenuItem menuItem = edgeCMF.createMenuItem(networkView, ev);
			addCyMenuItem(ev, menu, menuItem, tracker, viewTaskFactoryListener.cyEdgeViewContextMenuFactory.get(edgeCMF));
		}
		
		return menu;
	}
	
	public JPopupMenu createNodeMenu(View<CyNode> nodeView, 
			CyNetworkView networkView, 
			VisualLexicon visualLexicon,
			ViewTaskFactoryListener viewTaskFactoryListener) {
		
		Collection<NodeViewTaskFactory> usableTFs = getPreferredActions(viewTaskFactoryListener.nodeViewTFs, "NEW");
		Collection<CyNodeViewContextMenuFactory> usableCMFs = getPreferredActions(viewTaskFactoryListener.cyNodeViewContextMenuFactory, "NEW");
		View<CyNode> nv = nodeView;
		
		usableTFs.addAll(getPreferredActions(viewTaskFactoryListener.nodeViewTFs,"Edge"));
		usableCMFs.addAll(getPreferredActions(viewTaskFactoryListener.cyNodeViewContextMenuFactory,"Edge"));
		
		String nodeLabel = networkView.getModel().getRow(nv.getModel()).get(CyNetwork.NAME, String.class);
		JPopupMenu menu = createMenu(nodeLabel);
		JMenuTracker tracker = new JMenuTracker(menu);

		for (NodeViewTaskFactory nvtf : usableTFs) {
			Object context = null;
			NamedTaskFactory provisioner = factoryProvisioner.createFor(nvtf, nv, networkView);
			addMenuItem(nv, menu, provisioner, context, tracker, viewTaskFactoryListener.nodeViewTFs.get(nvtf), this.visualLexicon);
		}

		for (CyNodeViewContextMenuFactory nodeCMF: usableCMFs) {
			// menu.add(nodeVMF.createMenuItem(m_view, nv).getMenuItem());
			CyMenuItem menuItem = nodeCMF.createMenuItem(networkView, nv);
			addCyMenuItem(nv, menu, menuItem, tracker, viewTaskFactoryListener.cyNodeViewContextMenuFactory.get(nodeCMF));
		}
		
		return menu;
	}
	
	private <T> Collection<T> getPreferredActions(Map<T,Map> tfs, String action) {
		// if the action is null, return all available
		if ( action == null ) {
			return tfs.keySet();
		}

		// otherwise figure out if any TaskFactories match the specified preferred action
		java.util.List<T> usableTFs = new ArrayList<T>();
		for ( T evtf : tfs.keySet() ) {
			String prefAction = (String)(tfs.get( evtf ).get(PREFERRED_ACTION));
			// assume action is NEW if no action specified
			if (prefAction == null)
				prefAction = "NEW";
			if ( action.equalsIgnoreCase(prefAction) )
				usableTFs.add(evtf);
		}
		return usableTFs;
	}
	
	public JPopupMenu createNetworkMenu(CyNetworkView networkView, VisualLexicon visualLexicon,
			ViewTaskFactoryListener viewTaskFactoryListener) {
		
		Collection<NetworkViewTaskFactory> usableTFs = getPreferredActions(viewTaskFactoryListener.emptySpaceTFs, "NEW");
		Collection<NetworkViewLocationTaskFactory> usableTFs2 = getPreferredActions(viewTaskFactoryListener.networkViewLocationTFs, "NEW");
		Collection<CyNetworkViewContextMenuFactory> usableCMFs = getPreferredActions(viewTaskFactoryListener.cyNetworkViewContextMenuFactory, "NEW");
		
		final JPopupMenu menu = createMenu("Double Click Menu: empty");
		final JMenuTracker tracker = new JMenuTracker(menu);
		
		for ( NetworkViewTaskFactory nvtf : usableTFs ) {
			NamedTaskFactory provisioner = factoryProvisioner.createFor(nvtf, networkView);
			addMenuItem(null, menu, provisioner, null, tracker, viewTaskFactoryListener.emptySpaceTFs.get( nvtf ), this.visualLexicon);
		}
		
		for ( NetworkViewLocationTaskFactory nvltf : usableTFs2 ) {
//			NamedTaskFactory provisioner = factoryProvisioner.createFor(nvltf, networkView, rawPt, xformPt);
//			addMenuItem(null, menu, provisioner, null, tracker, viewTaskFactoryListener.networkViewLocationTFs.get( nvltf ), this.visualLexicon);
		}
		
		for (CyNetworkViewContextMenuFactory netVMF: usableCMFs) {
			CyMenuItem menuItem = netVMF.createMenuItem(networkView);
			addCyMenuItem(networkView, menu, menuItem, tracker, viewTaskFactoryListener.cyNetworkViewContextMenuFactory.get(netVMF));
		}
		
		return menu;
	}
	
	public JPopupMenu createNetworkLocationMenu(CyNetworkView networkView, VisualLexicon visualLexicon,
			Map<NetworkViewLocationTaskFactory, Map<String, Object>> taskFactories) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuTracker tracker = new JMenuTracker(popupMenu);
		
		for (Entry<NetworkViewLocationTaskFactory, Map<String, Object>> entry : taskFactories.entrySet()) {

			NetworkViewLocationTaskFactory networkViewLocationTaskFactory = entry.getKey();
			Map<String, Object> properties = entry.getValue();
			
			// TODO: Ding implementation for the NetworkViewLocation Task objects also takes 2D coordinates for 
			// the source of the event. Determine if appropriate to support NetworkViewLocation Tasks.
			
			/*
			TaskFactory taskFactory = taskFactoryProvider.createFor(networkViewLocationTaskFactory, networkView);
				createMenuItem(null, visualLexicon, popupMenu, taskFactory, null, tracker, properties);
			*/
		}
		
		return popupMenu;
	}
	
	private void createMenuItem(
			View<?> view,
			JPopupMenu popup,
			NamedTaskFactory tf,
			Object tunableContext,
	        JMenuTracker tracker,
	        Map props,
	        VisualLexicon visualLexicon) {

		String title = (String)(props.get(TITLE));
		String pref = (String)(props.get(PREFERRED_MENU));
		String toolTip = (String) (props.get(TOOLTIP));
		String menuGravity = (String) (props.get(MENU_GRAVITY));
		String prefAction = (String)(props.get(PREFERRED_ACTION));
		boolean insertSepBefore = getBooleanProperty(props, INSERT_SEPARATOR_BEFORE);
		boolean insertSepAfter = getBooleanProperty(props, INSERT_SEPARATOR_AFTER);
		boolean useCheckBoxMenuItem = getBooleanProperty(props, "useCheckBoxMenuItem");
		double gravity;

		if (menuGravity != null) {
			gravity = Double.parseDouble(menuGravity);
		} else  {
			//gravity = largeValue++;
			gravity = -1;  // Alphabetize by default
		}

		if(pref == null) {
			if(prefAction != null && prefAction.equalsIgnoreCase("OPEN"))
				pref = ".";
			else
				pref = APPS_MENU;
		}
	
		// otherwise create our own popup menus 
		final Object targetVisualProperty = props.get("targetVP");
		boolean isSelected = false;
		if(view != null) {
			if (targetVisualProperty != null && targetVisualProperty instanceof String ) {
				// TODO remove this at first opportunity whenever lookup gets refactored. 
				Class<?> clazz = CyNetwork.class;
				if ( view.getModel() instanceof CyNode )
					clazz = CyNode.class;
				else if ( view.getModel() instanceof CyEdge )
					clazz = CyEdge.class;

				final VisualProperty<?> vp = visualLexicon.lookup(clazz, targetVisualProperty.toString());
				if (vp == null)
					isSelected = false;
				else
					isSelected = view.isValueLocked(vp);
			} else if ( targetVisualProperty instanceof VisualProperty )
				isSelected = view.isValueLocked((VisualProperty<?>)targetVisualProperty);
		}

		// no title
		if ( title == null ) {
			int last = pref.lastIndexOf(".");

			// if the preferred menu is delimited
			if (last > 0) {
				title = pref.substring(last + 1);
				pref = pref.substring(0, last);
				// System.out.println("no title, pref = "+pref);
				final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
				final JMenuItem item = createMenuItem(tf, title,useCheckBoxMenuItem, toolTip);
				if (useCheckBoxMenuItem) {
					final JCheckBoxMenuItem checkBox = (JCheckBoxMenuItem)item; 
					checkBox.setSelected(isSelected);
				}
				if (insertSepBefore)
					gravityTracker.addMenuSeparator(gravity-.0001);
				gravityTracker.addMenuItem(item, gravity);
				if (insertSepAfter)
					gravityTracker.addMenuSeparator(gravity+.0001);
			// otherwise just use the preferred menu as the menuitem name
			} else {
				title = pref;
				// popup.add( createMenuItem(tf, title, useCheckBoxMenuItem, toolTip) );
				final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
				final JMenuItem item = createMenuItem(tf, title, useCheckBoxMenuItem, toolTip);
				gravityTracker.addMenuItem(item, gravity);
			}

		// title and preferred menu
		} else {
			// System.out.println("title, pref = "+pref);
			final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
			if (insertSepBefore)
				gravityTracker.addMenuSeparator(gravity-.0001);
			gravityTracker.addMenuItem(createMenuItem(tf, title,useCheckBoxMenuItem, toolTip), gravity);
			if (insertSepAfter)
				gravityTracker.addMenuSeparator(gravity+.0001);
		}
	}
	
	private void addMenuItem(
			View<?> view,
			JPopupMenu popup,
			NamedTaskFactory tf,
			Object tunableContext,
	        JMenuTracker tracker,
	        Map props,
	        VisualLexicon visualLexicon) {

		String title = (String)(props.get(TITLE));
		String pref = (String)(props.get(PREFERRED_MENU));
		String toolTip = (String) (props.get(TOOLTIP));
		String menuGravity = (String) (props.get(MENU_GRAVITY));
		String prefAction = (String)(props.get(PREFERRED_ACTION));
		boolean insertSepBefore = getBooleanProperty(props, INSERT_SEPARATOR_BEFORE);
		boolean insertSepAfter = getBooleanProperty(props, INSERT_SEPARATOR_AFTER);
		boolean useCheckBoxMenuItem = getBooleanProperty(props, "useCheckBoxMenuItem");
		double gravity;

		if (menuGravity != null) {
			gravity = Double.parseDouble(menuGravity);
		} else  {
			//gravity = largeValue++;
			gravity = -1;  // Alphabetize by default
		}

		if(pref == null) {
			if(prefAction != null && prefAction.equalsIgnoreCase("OPEN"))
				pref = ".";
			else
				pref = APPS_MENU;
		}
	
		// otherwise create our own popup menus 
		final Object targetVisualProperty = props.get("targetVP");
		boolean isSelected = false;
		if(view != null) {
			if (targetVisualProperty != null && targetVisualProperty instanceof String ) {
				// TODO remove this at first opportunity whenever lookup gets refactored. 
				Class<?> clazz = CyNetwork.class;
				if ( view.getModel() instanceof CyNode )
					clazz = CyNode.class;
				else if ( view.getModel() instanceof CyEdge )
					clazz = CyEdge.class;

				final VisualProperty<?> vp = visualLexicon.lookup(clazz, targetVisualProperty.toString());
				if (vp == null)
					isSelected = false;
				else
					isSelected = view.isValueLocked(vp);
			} else if ( targetVisualProperty instanceof VisualProperty )
				isSelected = view.isValueLocked((VisualProperty<?>)targetVisualProperty);
		}

		// no title
		if ( title == null ) {
			int last = pref.lastIndexOf(".");

			// if the preferred menu is delimited
			if (last > 0) {
				title = pref.substring(last + 1);
				pref = pref.substring(0, last);
				// System.out.println("no title, pref = "+pref);
				final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
				final JMenuItem item = createMenuItem(tf, title,useCheckBoxMenuItem, toolTip);
				if (useCheckBoxMenuItem) {
					final JCheckBoxMenuItem checkBox = (JCheckBoxMenuItem)item; 
					checkBox.setSelected(isSelected);
				}
				if (insertSepBefore)
					gravityTracker.addMenuSeparator(gravity-.0001);
				gravityTracker.addMenuItem(item, gravity);
				if (insertSepAfter)
					gravityTracker.addMenuSeparator(gravity+.0001);
			// otherwise just use the preferred menu as the menuitem name
			} else {
				title = pref;
				// popup.add( createMenuItem(tf, title, useCheckBoxMenuItem, toolTip) );
				final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
				final JMenuItem item = createMenuItem(tf, title, useCheckBoxMenuItem, toolTip);
				gravityTracker.addMenuItem(item, gravity);
			}

		// title and preferred menu
		} else {
			// System.out.println("title, pref = "+pref);
			final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
			if (insertSepBefore)
				gravityTracker.addMenuSeparator(gravity-.0001);
			gravityTracker.addMenuItem(createMenuItem(tf, title,useCheckBoxMenuItem, toolTip), gravity);
			if (insertSepAfter)
				gravityTracker.addMenuSeparator(gravity+.0001);
		}
	}
	
	private void addCyMenuItem(View<?> view, JPopupMenu popup, CyMenuItem menuItem, JMenuTracker tracker, Map props) {
		String pref = null;
		if (props != null)
			pref = (String)(props.get(PREFERRED_MENU));

		if (pref == null)
			pref = APPS_MENU;

		// This is a *very* special case we used to help with Dynamic Linkout
		if (pref.equalsIgnoreCase(menuItem.getMenuItem().getText()) && menuItem.getMenuItem() instanceof JMenu) {
			final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
			JMenu menu = (JMenu)menuItem.getMenuItem();
			for (int menuIndex = 0; menuIndex < menu.getItemCount(); menuIndex++) {
				JMenuItem item = menu.getItem(menuIndex);
				gravityTracker.addMenuItem(item, -1);
			}
			return;
		}

		final GravityTracker gravityTracker = tracker.getGravityTracker(pref);
		gravityTracker.addMenuItem(menuItem.getMenuItem(), menuItem.getMenuGravity());
		return;
	}
	
	/**
 	 * Get a boolean property
 	 */
	private boolean getBooleanProperty(Map props, String property) {
		String value = (String)(props.get(property)); // get the property
		if (value == null || value.length() == 0) return false;
		try {
			return Boolean.parseBoolean(value);
		} catch (Exception e) { 
			return false; 
		}
	}
		
	private JMenuItem createMenuItem(TaskFactory tf, String title, boolean useCheckBoxMenuItem, String toolTipText) {
		JMenuItem item;
		PopupAction action = new PopupAction(tf, title);
		if ( useCheckBoxMenuItem )
			item = new JCheckBoxMenuItem(action);
		else
			item = new JMenuItem(action);

		item.setEnabled(tf.isReady());

		item.setToolTipText(toolTipText);
		return item;
	}
	
	private class PopupAction extends AbstractAction {
		TaskFactory tf;
		PopupAction(TaskFactory tf, String title) {
			super( title );
			this.tf = tf;
		}

		public void actionPerformed(ActionEvent ae) {
			taskManager.execute(tf.createTaskIterator());
		}
	}
	
	private JPopupMenu createMenu(String title) {
		final JPopupMenu menu = new JPopupMenu(title);
		menu.addPopupMenuListener(new PopupMenuListener() {
			@Override
			public void popupMenuWillBecomeVisible(PopupMenuEvent arg0) {
			}
			
			@Override
			public void popupMenuWillBecomeInvisible(PopupMenuEvent arg0) {
				menu.removeAll();
			}
			
			@Override
			public void popupMenuCanceled(PopupMenuEvent arg0) {
				menu.removeAll();
			}
		});
		return menu;
	}
}
